#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Utility functions for dealing with Vcs URLs of various types."""

__all__ = [
    "fixup_broken_git_url",
    "sanitize_url",
    "determine_browser_url",
]


import posixpath
import re
from typing import Callable, List, Optional, Union
from urllib.parse import urlparse, urlunparse

from debmutate.vcs import split_vcs_url, unsplit_vcs_url
from upstream_ontologist.vcs import (
    canonical_git_repo_url,
    convert_cvs_list_to_str,
    drop_vcs_in_scheme,
    find_public_repo_url,
    find_secure_repo_url,
    fixup_broken_git_details,
    fixup_rcp_style_git_repo_url,
    is_gitlab_site,
)


def find_public_vcs_url(url: str) -> Optional[str]:
    (repo_url, branch, subpath) = split_vcs_url(url)
    revised_url = find_public_repo_url(repo_url)
    if revised_url is not None:
        return unsplit_vcs_url(revised_url, branch, subpath)
    return None


def fixup_rcp_style_git_url(url: str) -> str:
    (repo_url, branch, subpath) = split_vcs_url(url)
    repo_url = fixup_rcp_style_git_repo_url(repo_url)
    return unsplit_vcs_url(repo_url, branch, subpath)


def determine_gitlab_browser_url(url: str) -> str:
    (url, branch, subpath) = split_vcs_url(url)
    parsed_url = urlparse(url.rstrip("/"))
    # TODO(jelmer): Add support for branches
    path = parsed_url.path
    if path.endswith(".git"):
        path = path[: -len(".git")]
    if subpath and not branch:
        branch = "HEAD"
    if branch:
        path = path + "/tree/%s" % branch
    if subpath:
        path = path + "/" + subpath
    return f"https://{parsed_url.hostname}{path}"


def determine_browser_url(vcs_type, vcs_url: str) -> Optional[str]:
    repo_url, branch, subpath = split_vcs_url(vcs_url)
    parsed = urlparse(repo_url.rstrip("/"))
    if is_gitlab_site(parsed.netloc):
        return determine_gitlab_browser_url(vcs_url)
    if parsed.netloc == "github.com":
        path = parsed.path
        if path.endswith(".git"):
            path = path[:-4]
        if subpath and not branch:
            branch = "HEAD"
        if branch:
            path = posixpath.join(path, "tree", branch)
        if subpath:
            path = posixpath.join(path, subpath)
        return urlunparse(
            ("https", parsed.netloc, path, parsed.query,
             parsed.params, parsed.fragment)
        )
    if (
        parsed.netloc in ("code.launchpad.net", "launchpad.net")
        and not branch
        and not subpath
    ):
        return urlunparse(
            (
                "https",
                "code.launchpad.net",
                parsed.path,
                parsed.query,
                parsed.params,
                parsed.fragment,
            )
        )
    if parsed.hostname in ("git.savannah.gnu.org", "git.sv.gnu.org"):
        path_elements = parsed.path.strip("/").split("/")
        if parsed.scheme == "https" and path_elements[0] == "git":
            path_elements.pop(0)
        # Why cgit and not gitweb?
        path_elements.insert(0, "cgit")
        return urlunparse(
            ("https", parsed.netloc, "/".join(path_elements), None, None, None)
        )
    if parsed.hostname in ("git.code.sf.net", "git.code.sourceforge.net"):
        path_elements = parsed.path.strip("/").split("/")
        if path_elements[0] != "p":
            return None
        project = path_elements[1]
        repository = path_elements[2]
        path_elements = ["p", project, repository]
        if branch is not None:
            path_elements.extend(["ci", branch, "tree"])
        elif subpath is not None:
            path_elements.extend(["ci", "HEAD", "tree"])
        if subpath is not None:
            path_elements.append(subpath)
        return urlunparse(
            ("https", "sourceforge.net", "/".join(path_elements), None, None,
             None)
        )
    return None


def canonicalize_vcs_browser_url(url: str) -> str:
    url = url.replace(
        "https://svn.debian.org/wsvn/", "https://anonscm.debian.org/viewvc/"
    )
    url = url.replace(
        "http://svn.debian.org/wsvn/", "https://anonscm.debian.org/viewvc/"
    )
    url = url.replace(
        "https://git.debian.org/?p=", "https://anonscm.debian.org/git/")
    url = url.replace(
        "http://git.debian.org/?p=", "https://anonscm.debian.org/git/")
    url = url.replace(
        "https://bzr.debian.org/loggerhead/",
        "https://anonscm.debian.org/loggerhead/"
    )
    url = url.replace(
        "http://bzr.debian.org/loggerhead/",
        "https://anonscm.debian.org/loggerhead/"
    )
    return re.sub(
        r"^https?://salsa.debian.org/([^/]+/[^/]+)\.git/?$",
        "https://salsa.debian.org/\\1",
        url,
    )


def canonical_vcs_git_url(url: str) -> str:
    (repo_url, branch, subpath) = split_vcs_url(url)
    repo_url = canonical_git_repo_url(repo_url)
    return unsplit_vcs_url(repo_url, branch, subpath)


canonicalize_vcs_fns = {
    "Browser": canonicalize_vcs_browser_url,
    "Git": canonical_vcs_git_url,
}


def canonicalize_vcs_url(vcs_type: str, url: str) -> str:
    try:
        return canonicalize_vcs_fns[vcs_type](url)
    except KeyError:
        return url


def find_secure_vcs_url(url: str, net_access: bool = True) -> Optional[str]:
    repo_url: Optional[str]
    (repo_url, branch, subpath) = split_vcs_url(url)
    repo_url = find_secure_repo_url(
        repo_url, branch=branch, net_access=net_access)
    if repo_url is None:
        return None

    return unsplit_vcs_url(repo_url, branch, subpath)


def fixup_broken_git_url(url: str) -> str:
    """Attempt to fix up broken Git URLs.

    A common misspelling is to add an extra ":" after the hostname
    """
    repo_url, branch, subpath = split_vcs_url(url)
    newrepo_url, newbranch, newsubpath = fixup_broken_git_details(
        repo_url, branch, subpath
    )
    if newrepo_url != repo_url or newbranch != branch or newsubpath != subpath:
        return unsplit_vcs_url(newrepo_url, newbranch, newsubpath)
    return url


SANITIZERS: List[Callable[[str], str]] = [
    drop_vcs_in_scheme,
    fixup_broken_git_url,
    fixup_rcp_style_git_url,
    lambda url: find_public_vcs_url(url) or url,
    canonical_vcs_git_url,
    lambda url: find_secure_vcs_url(url, net_access=False) or url,
]


def sanitize_url(url: Union[str, List[str]]) -> str:
    """Sanitize a version control URL."""
    if isinstance(url, list):
        url_str = convert_cvs_list_to_str(url)
    else:
        url_str = url
    url_str = url_str.strip()
    for sanitizer in SANITIZERS:
        url_str = sanitizer(url_str)
    return url_str



#!/usr/bin/python3

from email.utils import parseaddr

from lintian_brush.fixer import LintianIssue, control, report_result

QA_MAINTAINER = "Debian QA Group <packages@qa.debian.org>"

with control as updater:
    try:
        fullname, email = parseaddr(updater.source["Maintainer"])
    except KeyError:
        # No maintainer? Weird, but sure.
        pass
    else:
        if email == "packages@qa.debian.org":
            issue = LintianIssue(
                updater.source, 'faulty-debian-qa-group-phrase',
                f'Maintainer {fullname}')
            if issue.should_fix():
                updater.source["Maintainer"] = QA_MAINTAINER
                issue.report_fixed()


report_result("Fix Debian QA group name.")



#!/usr/bin/python3

import os
import sys

from debmutate.control import (
    add_dependency,
)
from debmutate.debhelper import (
    get_debhelper_compat_level,
)

from lintian_brush.fixer import control, fixed_lintian_tag, report_result

compat_version = get_debhelper_compat_level()
if compat_version is None or compat_version <= 11:
    # N/A
    sys.exit(0)


added = set()


# Add Pre-Depends: ${misc:Pre-Depends} iff:
# - a package has both a init script and a (ubuntu | systemd) unit

with control as updater:
    for binary in updater.binaries:
        name = binary["Package"]
        if not os.path.exists('debian/%s.init' % name):
            continue
        if not (os.path.exists('debian/%s.service' % name) or
                os.path.exists('debian/%s.upstart' % name)):
            continue
        if "${misc:Pre-Depends}" in binary.get("Pre-Depends", ""):
            continue
        binary["Pre-Depends"] = add_dependency(
            binary.get("Pre-Depends", ""),
            "${misc:Pre-Depends}")
        added.add(name)
        fixed_lintian_tag(
            updater.source, 'skip-systemd-native-flag-missing-pre-depends')


report_result(
    "Add missing Pre-Depends: ${misc:Pre-Depends} in %s." %
    ", ".join(sorted(added)))



#!/usr/bin/python
# Copyright (C) 2020 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Standards-Version handling."""

import json
import os
from datetime import datetime
from typing import Iterator, Tuple

from debmutate.control import parse_standards_version
from iso8601 import parse_date

from .lintian import LINTIAN_DATA_PATH

RELEASE_DATES_PATH = os.path.join(
    LINTIAN_DATA_PATH, "debian-policy/releases.json")
OLD_RELEASE_DATES_PATH = os.path.join(
    LINTIAN_DATA_PATH, "standards-version/release-dates")


def iter_standards_versions() -> Iterator[Tuple[Tuple[int, ...], datetime]]:
    try:
        with open(RELEASE_DATES_PATH) as f:
            data = json.load(f)
            for release in data['releases']:
                version = release['version']
                yield (
                    parse_standards_version(version),
                    parse_date(release['timestamp']))
    except FileNotFoundError:
        with open(OLD_RELEASE_DATES_PATH) as f:
            for line in f:
                if line.startswith("#") or not line.strip():
                    continue
                (version, ts) = line.split()
                yield (parse_standards_version(version),
                       datetime.fromtimestamp(int(ts)))


def latest_standards_version() -> str:
    return ".".join(map(str, next(iter_standards_versions())[0]))



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for lintian_brush fixers."""

import os
import shutil
import subprocess
import sys
import tempfile
import unittest

from debian.changelog import Changelog
from lintian_brush import (
    available_lintian_fixers,
    increment_version,
    parse_script_fixer_output,
    select_fixers,
)
from lintian_brush.lintian_overrides import (
    load_renamed_tags,
)


class FixerTestCase(unittest.TestCase):
    """Test case that runs a fixer test."""

    def __init__(self, fixer, fixer_name, name, path):
        self._fixer = fixer
        self._fixer_name = fixer_name
        self._test_name = name
        self._path = path
        self.maxDiff = None
        super().__init__()

    def setUp(self):
        self._tempdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self._tempdir)
        self._testdir = os.path.join(self._tempdir, "testdir")

        def ignore(src, names):
            return [name for name in names if name.endswith("~")]

        shutil.copytree(
            os.path.join(self._path, "in"), self._testdir, symlinks=True,
            ignore=ignore
        )

    def id(self):
        return f"{__name__}.{self._fixer_name}.{self._test_name}"

    def __str__(self):
        return f"fixer test: {self._test_name} for {self._fixer_name}"

    def runTest(self):
        xfail_path = os.path.join(self._path, "xfail")
        if os.path.exists(xfail_path):
            with open(xfail_path) as f:
                reason = f.read()  # noqa: F841
            unittest.expectedFailure(self)
            return
        env = {}
        for key in ['PATH']:
            if key in os.environ:
                env[key] = os.environ[key]
        cl_path = os.path.join(self._testdir, "debian/changelog")
        if os.path.exists(cl_path):
            with open(cl_path, "rb") as f:
                cl = Changelog(f, max_blocks=1)
            if cl.distributions == "UNRELEASED":
                current_version = cl.version
            else:
                current_version = cl.version
            current_version = increment_version(current_version)
        else:
            current_version = "1.0-1"
        env["CURRENT_VERSION"] = str(current_version)
        env["NET_ACCESS"] = "disallow"
        env["MINIMUM_CERTAINTY"] = "possible"
        env["PYTHONPATH"] = ':'.join(
            [os.path.dirname(os.path.dirname(os.path.dirname(__file__)))]
            + sys.path)
        env_path = os.path.join(self._path, "env")
        if os.path.exists(env_path):
            with open(env_path) as f:
                for line in f:
                    key, value = line.rstrip("\n").split("=")
                    env[key] = value
        p = subprocess.Popen(
            self._fixer.script_path, cwd=self._testdir, stdout=subprocess.PIPE,
            env=env
        )
        (stdout, err) = p.communicate(b"")
        self.assertEqual(p.returncode, 0)
        out_path = os.path.join(self._path, "out")
        p = subprocess.Popen(
            [
                "diff",
                "--no-dereference",
                "-x",
                "*~",
                "-ur",
                os.path.join(self._path, os.readlink(out_path))
                if os.path.islink(out_path)
                else out_path,
                self._testdir,
            ],
            stdout=subprocess.PIPE,
        )
        (diff, stderr) = p.communicate(b"")
        self.assertIn(
            p.returncode, (0, 1),
            "Unexpected exit code %d" % p.returncode)
        if diff.decode() != "":
            raise AssertionError("unexpected output: %s" % diff.decode())
        self.assertMultiLineEqual(diff.decode(), "")

        if (
            not os.path.islink(out_path)
            or os.readlink(os.path.join(self._path, "out")) != "in"
        ):
            check_message = True
            result = parse_script_fixer_output(stdout.decode())
            self.assertTrue(
                set(result.fixed_lintian_tags).issubset(
                    self._fixer.lintian_tags),
                "fixer {} claims to fix tags ({!r}) not declared "
                "in index.desc ({!r})".format(
                    self._fixer_name,
                    result.fixed_lintian_tags,
                    self._fixer.lintian_tags,
                ),
            )
        else:
            check_message = False

        message_path = os.path.join(self._path, "message")
        if os.path.exists(message_path) or check_message:
            with open(message_path) as f:
                # Assert that message on stdout matches
                self.assertEqual(stdout.decode().strip(), f.read().strip())


class SaneFixerTests(unittest.TestCase):
    """Check that the test is sensible."""

    def id(self):
        return f"{__name__}.{self.fixer.name}.sane"

    def __str__(self):
        return "fixer sanity test: %s" % (self.fixer.name)

    def __init__(self, fixer):
        self.fixer = fixer
        super().__init__()

    def runTest(self):
        self.assertTrue(
            os.path.exists(self.fixer.script_path),
            "Script %s missing" % self.fixer.script_path,
        )

        renames = load_renamed_tags()
        for tag in self.fixer.lintian_tags:
            self.assertNotIn(
                tag, renames,
                f"Tag {tag} has been renamed to {renames.get(tag)}"
            )


def iter_test_cases(fixer):
    test_dir = os.path.join(os.path.dirname(__file__), "..", "..", "tests")
    testpath = os.path.join(test_dir, fixer.name)
    if not os.path.isdir(testpath):
        return
    for testname in os.listdir(testpath):
        yield FixerTestCase(
            fixer_name=fixer.name,
            fixer=fixer,
            name=testname,
            path=os.path.join(test_dir, fixer.name, testname),
        )


def test_suite():
    suite = unittest.TestSuite()
    fixers = available_lintian_fixers()
    for fixer in fixers:
        for test_case in iter_test_cases(fixer):
            suite.addTest(test_case)
        suite.addTest(SaneFixerTests(fixer))
    return suite


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--fixer",
        type=str,
        default=None,
        action="append",
        help="Fixer for which to run tests.",
    )
    parser.add_argument(
        "--exclude", type=str, action="append", help="Exclude a fixer.")
    args = parser.parse_args()

    fixers = list(available_lintian_fixers())
    if args.fixer:
        try:
            fixers = select_fixers(
                fixers, names=args.fixer, exclude=args.exclude)
        except KeyError as e:
            print("Selected fixer %s does not exist." % (
                e.args[0]), file=sys.stderr)
            sys.exit(0)

    suite = unittest.TestSuite()
    for fixer in fixers:
        for test_case in iter_test_cases(fixer):
            suite.addTest(test_case)

    runner = unittest.TextTestRunner()
    runner.run(suite)



#!/usr/bin/python3

import os
import sys

from debian.copyright import (
    Copyright,
    FilesParagraph,
    License,
    LicenseParagraph,
)
from lintian_brush.fixer import (
    LintianIssue,
    meets_minimum_certainty,
    report_result,
)

CERTAINTY = 'possible'

if not meets_minimum_certainty(CERTAINTY):
    sys.exit(0)

if os.path.exists('debian/copyright'):
    sys.exit(0)


try:
    import decopy  # noqa: F401
except ModuleNotFoundError:
    # No decopy
    sys.exit(2)

from decopy.cmdoptions import process_options  # noqa: E402
from decopy.datatypes import License as DecopyLicense  # noqa: E402
from decopy.dep5 import Copyright as DecopyCopyright  # noqa: E402
from decopy.dep5 import Group  # noqa: E402
from decopy.tree import DirInfo, RootInfo  # noqa: E402

options = process_options([
    '--root=.',
    '--no-progress',
    '--mode=full',
    '--output=debian/copyright',
    ])

filetree = RootInfo.build(options)
copyright_ = DecopyCopyright.build(filetree, options)

copyright_.process(filetree)
filetree.process(options)

groups = copyright_.get_group_dict(options)

for fileinfo in filetree:
    if fileinfo.group:
        continue
    if isinstance(fileinfo, DirInfo):
        continue

    file_key = fileinfo.get_group_key(options)
    group = groups.setdefault(file_key, Group(file_key))
    group.add_file(fileinfo)
    fileinfo.group = group

licenses = set()

c = Copyright()
# Print files paragraphs
for _, group in sorted(
        groups.items(), key=lambda i: i[1].sort_key(options)):

    if not group.copyright_block_valid():
        continue

    licenses.update(group.licenses.keys())

    if options.glob:
        files = group.files.get_patterns()
    else:
        files = group.files.sorted_members()

    if group.copyrights:
        holders = '\n           '.join(group.copyrights.sorted_members())
    else:
        holders = 'Unknown'
    files_paragraph = FilesParagraph.create(
        list(files), holders, License(group.license))

    comments = group.get_comments()
    if comments:
        files_paragraph.comment = comments

    c.add_files_paragraph(files_paragraph)

# Print license paragraphs
for key in sorted(licenses):
    license_ = DecopyLicense.get(key)
    license_paragraph = LicenseParagraph.create(
        License(license_.name))
    license_paragraph.comment = (
        "Add the corresponding license text here")  # type: ignore
    c.add_license_paragraph(license_paragraph)


issue = LintianIssue('source', 'no-copyright-file')
if issue.should_fix():
    with open('debian/copyright', 'w') as f:
        c.dump(f)
    issue.report_fixed()

report_result('Create a debian/copyright file.', certainty=CERTAINTY)



#!/usr/bin/python3

from lintian_brush.fixer import LintianIssue, control, report_result
from lintian_brush.vcs import canonicalize_vcs_url

fields = set()


with control as updater:
    for name in updater.source:
        if not name.startswith("Vcs-"):
            continue
        new_value = canonicalize_vcs_url(
            name[len("Vcs-"):], updater.source[name])
        if new_value != updater.source[name]:
            issue = LintianIssue(
                updater.source, 'vcs-field-not-canonical',
                f'{updater.source[name]} {new_value}')
            if issue.should_fix():
                issue.report_fixed()
                updater.source[name] = new_value
                fields.add(name)

report_result("Use canonical URL in " + ', '.join(sorted(fields)) + '.')



#!/usr/bin/python3

import os

from lintian_brush.fixer import fixed_lintian_tag, report_result

removed = []

for name in os.listdir('debian'):
    if name.endswith('.linda-overrides'):
        os.unlink(os.path.join('debian', name))
        removed.append(name)
        fixed_lintian_tag(
            'source', 'package-contains-linda-override',
            'usr/share/linda/overrides/%s' % name[:-len('.linda-overrides')])

report_result('Remove obsolete linda overrides: ' + ', '.join(removed))



#!/usr/bin/python3

from debmutate.deb822 import Deb822Editor

from lintian_brush.fixer import LintianIssue, report_result

try:
    with Deb822Editor(path='debian/copyright') as updater:
        for paragraph in updater.paragraphs:
            if 'Files' not in paragraph:
                continue
            if ',' not in paragraph['Files']:
                continue
            if '{' in paragraph['Files']:
                # Bash-style expansion?
                continue
            issue = LintianIssue(
                'source', 'comma-separated-files-in-dep5-copyright',
                'paragraph at line XX')
            if issue.should_fix():
                paragraph['Files'] = '\n' + '\n'.join(
                    ' ' + entry.strip()
                    for entry in paragraph['Files'].split(','))
                issue.report_fixed()

except FileNotFoundError:
    pass
except ValueError:
    # Not a machine-readable copyright file
    pass

report_result(
    "debian/copyright: Replace commas with whitespace to separate items "
    "in Files paragraph.")



#!/usr/bin/python3

from debmutate.watch import WatchEditor

from lintian_brush.fixer import (
    current_package_version,
    net_access_allowed,
    report_result,
    source_package_name,
)
from lintian_brush.watch import fix_github_releases, watch_entries_certainty

try:
    with WatchEditor() as updater:
        entries = fix_github_releases(updater)
        if net_access_allowed():
            certainty = watch_entries_certainty(
                entries, source_package_name(),
                expected_versions=[current_package_version().upstream_version])
        else:
            certainty = None
except FileNotFoundError:
    pass
else:
    report_result(
        "debian/watch: Use GitHub /tags rather than /releases page.",
        certainty=certainty)



#!/usr/bin/python3

import re
import sys

from debmutate.changelog import ChangelogEditor, distribution_is_unreleased

from debian.changelog import Version
from lintian_brush.fixer import LintianIssue, report_result

with ChangelogEditor() as editor:
    if len(editor.changelog) != 1:
        sys.exit(0)

    block = editor.changelog[-1]

    if not distribution_is_unreleased(block.distributions):
        sys.exit(0)

    if not re.fullmatch("2[0-9]{7}", block.version.upstream_version):
        sys.exit(0)

    issue = LintianIssue(
        "source", "new-package-uses-date-based-version-number", None)
    if issue.should_fix():
        block.version = Version("0~%s" % block.version)
        issue.report_fixed()

report_result("Use version prefix for date-based versionioning.")



#!/usr/bin/python3
import os
import re
import shlex
import sys
from contextlib import suppress

from debmutate._rules import (
    Makefile,
    check_cdbs,
    dh_invoke_drop_argument,
    dh_invoke_drop_with,
    dh_invoke_replace_argument,
    update_rules,
)
from debmutate.control import (
    ensure_exact_version,
    ensure_minimum_version,
    get_relation,
)
from debmutate.debhelper import (
    MaintscriptEditor,
    MaintscriptRemoveConffile,
    read_debhelper_compat_file,
)

from lintian_brush.debhelper import (
    DEBHELPER_BUILD_STEPS,
    detect_debhelper_buildsystem,
    lowest_non_deprecated_compat_level,
    maximum_debhelper_compat_version,
)
from lintian_brush.fixer import (
    compat_release,
    control,
    current_package_version,
    fixed_lintian_tag,
    is_debcargo_package,
    report_result,
    warn,
)

new_debhelper_compat_version = maximum_debhelper_compat_version(
    compat_release())

uses_cdbs = check_cdbs()

if uses_cdbs:
    # cdbs doesn't appear to support debhelper 11 or 12 just yet..
    new_debhelper_compat_version = min(new_debhelper_compat_version, 10)


# If the package uses autoconf, configure doesn't contain --runstatedir, and
# specifies --without autoreconf, then we can't upgrade beyond debhelper 10. If
# we do, configure fails because debhelper >= 11 specifies --runstatedir.

# We could choose to drop the --without autoreconf, but we don't know why the
# maintainer chose to specify it.
def autoreconf_disabled():
    try:
        mf = Makefile.from_path('debian/rules')
        for line in mf.dump_lines():
            if re.findall(b'--without.*autoreconf', line):
                return True

        # Another way to disable dh_autoreconf is to add an empty override
        # rule.
        for rule in mf.iter_rules(b'override_dh_autoreconf'):
            if rule.commands():
                return False
        else:
            return True
    except FileNotFoundError:
        return False
    return False


if autoreconf_disabled():
    with suppress(IsADirectoryError, FileNotFoundError), \
            open('configure', 'rb') as f:
        for line in f:
            if b'runstatedir' in line:
                break
        else:
            new_debhelper_compat_version = min(
                new_debhelper_compat_version, 10)
            warn(
                'Not upgrading beyond debhelper %d, since the package '
                'disables autoreconf but its configure does not provide '
                '--runstatedir.' % new_debhelper_compat_version)


if os.path.exists('debian/compat'):
    # Package currently stores compat version in debian/compat..

    current_debhelper_compat_version = read_debhelper_compat_file(
        'debian/compat')
    if current_debhelper_compat_version < new_debhelper_compat_version:
        with open('debian/compat', 'w') as cf:
            cf.write('%s\n' % new_debhelper_compat_version)
    else:
        # Nothing to do
        sys.exit(2)

    with control as updater:
        updater.source["Build-Depends"] = ensure_minimum_version(
            updater.source.get("Build-Depends", ""),
            "debhelper",
            "%d~" % new_debhelper_compat_version)
else:
    try:
        # Assume that the compat version is set in Build-Depends
        with control as updater:
            try:
                offset, debhelper_compat_relation = get_relation(
                    updater.source.get("Build-Depends", ""),
                    "debhelper-compat")
            except KeyError:
                sys.exit(2)
            else:
                if len(debhelper_compat_relation) > 1:
                    # Not sure how to deal with this..
                    sys.exit(2)
                if debhelper_compat_relation[0].version[0] != '=':
                    # Not sure how to deal with this..
                    sys.exit(2)
                current_debhelper_compat_version = int(
                    debhelper_compat_relation[0].version[1])
            if current_debhelper_compat_version < new_debhelper_compat_version:
                updater.source["Build-Depends"] = ensure_exact_version(
                        updater.source["Build-Depends"],
                        "debhelper-compat",
                        "%d" % new_debhelper_compat_version)
    except FileNotFoundError:
        # debcargo just uses the latest version and doesn't store debhelper
        # version explicitly.
        if is_debcargo_package():
            sys.exit(0)
        raise


# For a list of behavior changes between debhelper compat versions, see
# https://manpages.debian.org/testing/debhelper/debhelper.7.en.html#Supported_compatibility_levels


subitems = set()


def line_matches_command(target, line, command):
    if command is None:
        # Whatever
        return True

    if line.startswith(command + b' ') or line == command:
        return True

    return (target == (b'override_' + command) and
            line.startswith(b'$(overridden_command)'))


def update_line(line, orig, new, description):
    newline = line.replace(orig, new)
    if newline != line:
        subitems.add(description)
        changed = True
    else:
        changed = False
    return newline, changed


def update_line_drop_argument(target, line, command, argument, description):
    if line_matches_command(target, line, command) and argument in line:
        line = dh_invoke_drop_argument(line, argument)
        subitems.add(description)
        return line, True
    return line, False


def update_line_replace_argument(line, old, new, description):
    newline = dh_invoke_replace_argument(line, old, new)
    if newline != line:
        subitems.add(description)
        return newline, True
    return line, False


class PybuildUpgrader:

    def __init__(self):
        # Does the dh line specify --buildsystem=pybuild?
        self.upgraded = False

    def fix_line(self, line, target):
        """Upgrade from python_distutils to pybuild."""
        line, changed = update_line(
            line, b'--buildsystem=python_distutils', b'--buildsystem=pybuild',
            'Replace python_distutils buildsystem with pybuild.')
        line, changed = update_line(
            line, b'--buildsystem python_distutils', b'--buildsystem=pybuild',
            'Replace python_distutils buildsystem with pybuild.')
        line, changed = update_line(
            line, b'-O--buildsystem=python_distutils',
            b'-O--buildsystem=pybuild',
            'Replace python_distutils buildsystem with pybuild.')
        if target.decode() in DEBHELPER_BUILD_STEPS:
            step = target.decode()
        elif target.startswith(b'override_dh_auto_'):
            step = target[len(b'override_dh_auto_'):].decode()
        else:
            step = None
        if line.startswith(b'dh '):
            if b'buildsystem' not in line:
                buildsystem = detect_debhelper_buildsystem(step)
                if buildsystem == 'python_distutils':
                    line += b' --buildsystem=pybuild'
                    subitems.add(
                        'Replace python_distutils buildsystem with pybuild.')
                    self.upgraded = True
            else:
                if b'buildsystem=pybuild' in line:
                    self.upgraded = True
        if (line.startswith(b'dh_auto_') and
            b' -- ' in line and
            (self.upgraded or
                re.match(b'--buildsystem[= ]pybuild', line) or
                detect_debhelper_buildsystem(step) == 'pybuild')):
            line, rest = line.split(b' -- ', 1)
            if step is None:
                step = line[len(b'dh_auto_'):].split(b' ', 1)[0].decode()
            subitems.add('Replace python_distutils buildsystem with pybuild.')
            line = (b'PYBUILD_' + step.upper().encode() + b'_ARGS=' +
                    shlex.quote(rest.decode()).encode() + b' ' + line)

        return line


def upgrade_to_dh_prep(line, target):
    """Replace 'dh_clean -k' with 'dh_prep."""
    line, changed = update_line(
        line, b'dh_clean -k', b'dh_prep',
        'debian/rules: Replace dh_clean -k with dh_prep.')
    return line


class DhMissingUpgrader:
    """Replace --list-missing / --fail-missing with dh_missing."""

    def __init__(self):
        self.need_override_missing = False

    def fix_line(self, line, target):
        for arg in [b'--list-missing', b'-O--list-missing']:
            for command in [b'dh_install', b'dh']:
                line, changed = update_line_drop_argument(
                    target, line, command, arg,
                    'debian/rules: Rely on default use of dh_missing rather '
                    'than using dh_install --list-missing.')
        for arg in [b'--fail-missing', b'-O--fail-missing']:
            for command in [b'dh_install', b'dh']:
                line, changed = update_line_drop_argument(
                    target, line, command, arg,
                    'debian/rules: Move --fail-missing argument to dh_missing.'
                    )
                if changed:
                    if target == b'override_dh_install' or command == b'dh':
                        self.need_override_missing = True
                    else:
                        subitems.add(
                            'debian/rules: Move --fail-missing argument '
                            'to dh_missing.')
                        return [line, b'dh_missing --fail-missing']
        return line

    def fix_makefile(self, mf):
        if not self.need_override_missing:
            return
        try:
            [rule] = list(mf.iter_rules(b'override_dh_missing'))
        except ValueError:
            rule = mf.add_rule(b'override_dh_missing')
            rule.append_command(b'dh_missing --fail-missing')
        else:
            for i, line in enumerate(rule.lines):
                if line.startswith(b'dh_missing '):
                    if b'--fail-missing' in line:
                        return
                    else:
                        rule.lines[i] += b' --fail-missing'
            else:
                raise Exception(
                    'override_dh_missing exists, but has no call to '
                    'dh_missing')


def replace_deprecated_same_arch(line, target):
    if not line.startswith(b'dh'):
        return line
    line, _ = update_line_replace_argument(
        line, b'-s', b'-a', 'Replace deprecated -s with -a.')
    line, _ = update_line_replace_argument(
        line, b'--same-arch', b'--arch',
        'Replace deprecated --same-arch with --arch.')
    return line


def upgrade_to_no_stop_on_upgrade(line, target):
    if line.startswith(b'dh ') or line.startswith(b'dh_installinit'):
        line, changed = update_line(
            line, b'--no-restart-on-upgrade',
            b'--no-stop-on-upgrade',
            'Replace --no-restart-on-upgrade with --no-stop-on-upgrade.')
    return line


def debhelper_argument_order(line, target):
    if line.startswith(b'dh '):
        args = line.split(b' ')
        for possible_va in [b'$*', b'$@', b'${@}']:
            try:
                x = args.index(possible_va)
            except ValueError:
                continue
            break
        else:
            return line
        val = args.pop(x)
        args.insert(1, val)
        return b' '.join(args)
    return line


def override_dh_auto_test_drop_options(rule):
    if b'override_dh_auto_test' not in rule.targets:
        return
    if (len(rule.lines) < 2
            or rule.lines[1]
            != b'ifeq (,$(filter nocheck,$(DEB_BUILD_OPTIONS)))'
            or rule.lines[-1] != b'endif'):
        return
    subitems.add(
        'Drop check for DEB_BUILD_OPTIONS containing "nocheck", '
        'since debhelper now does this.')
    del rule.lines[1]
    del rule.lines[-1]


def upgrade_to_debhelper_12():

    pybuild_upgrader = PybuildUpgrader()
    dh_missing_upgrader = DhMissingUpgrader()
    update_rules([
        debhelper_argument_order,
        replace_deprecated_same_arch,
        pybuild_upgrader.fix_line,
        upgrade_to_dh_prep,
        upgrade_to_no_stop_on_upgrade,
        dh_missing_upgrader.fix_line,
        ], makefile_cb=dh_missing_upgrader.fix_makefile)


def upgrade_to_installsystemd(line, target):
    line = dh_invoke_drop_with(line, b'systemd')
    if line.startswith(b'dh_systemd_enable'):
        line, changed = update_line(
            line, b'dh_systemd_enable', b'dh_installsystemd',
            'Use dh_installsystemd rather than deprecated '
            'dh_systemd_enable.')
    if line.startswith(b'dh_systemd_start'):
        line, changed = update_line(
            line, b'dh_systemd_start', b'dh_installsystemd',
            'Use dh_installsystemd rather than deprecated '
            'dh_systemd_start.')
    return line


def rename_installsystemd_target(rule):
    rule.rename_target(
        b'override_dh_systemd_enable', b'override_dh_installsystemd')
    rule.rename_target(
        b'override_dh_systemd_start', b'override_dh_installsystemd')


def upgrade_to_debhelper_10():

    # dh_installinit will no longer install a file named debian/package as an
    # init script.

    with control as updater:
        for binary in updater.binaries:
            name = binary['Package']
            if os.path.isfile(os.path.join('debian', name)):
                os.rename(os.path.join('debian', name),
                          os.path.join('debian', name + '.init'))
                subitems.add(
                    f'Rename debian/{name} to debian/{name}.init.')


def upgrade_to_debhelper_11():

    update_rules(
        [upgrade_to_installsystemd], rule_cb=rename_installsystemd_target)
    for name in os.listdir('debian'):
        parts = name.split('.')
        if len(parts) < 2 or parts[-1] != 'upstart':
            continue
        if len(parts) == 3:
            package = parts[0]
            service = parts[1]
        elif len(parts) == 2:
            package = service = parts[0]
        os.unlink(os.path.join('debian', name))
        subitems.add('Drop obsolete upstart file %s.' % name)
        with MaintscriptEditor('debian/%s.maintscript' % package) as e:
            e.append(MaintscriptRemoveConffile(
                conffile='/etc/init/%s.conf' % service,
                prior_version=current_package_version(),
                package=None))


def drop_dh_missing_fail(line, target):
    """Remove dh_missing --fail-missing."""
    # Note that this is not currently replacing plain calls to dh_missing with
    # empty an "override_dh_missing" rule.

    if not line.startswith(b'dh_missing '):
        return line
    argv = line.split(b' ')
    for arg in [b'--fail-missing', b'-O--fail-missing']:
        if arg in argv:
            # If we added the --fail-missing argument first, then
            # don't announce that here.
            try:
                subitems.remove(
                    'debian/rules: Move --fail-missing argument '
                    'to dh_missing.')
            except KeyError:
                subitems.add(
                    'debian/rules: Drop --fail-missing argument '
                    'to dh_missing, which is now the default.')
            else:
                subitems.add(
                    'debian/rules: Drop --fail-missing argument, '
                    'now the default.')
            argv.remove(arg)
    return b' '.join(argv)


def upgrade_to_debhelper_13():

    # dh_installtempfiles will handle d/tmpfile{,s}.  It prefer the latter
    # and warns about the former.

    with control as updater:
        for binary in updater.binaries:
            name = binary['Package']
            tmpfile = os.path.join('debian', name + ".tmpfile")
            if os.path.isfile(tmpfile):
                os.rename(tmpfile, tmpfile + 's')
                subitems.add(
                    f'Rename {tmpfile} to {tmpfile}s.')
        if os.path.isfile(os.path.join('debian', "tmpfile")):
            tmpfile = os.path.join('debian', "tmpfile")
            os.rename(tmpfile, tmpfile + 's')
            subitems.add(
                f'Rename {tmpfile} to {tmpfile}s.')

    update_rules([
        # The dh_missing command will now default to --fail-missing.  This can
        # be reverted to a non-fatal warning by explicitly passing
        # --list-missing like it was in compat 12.
        drop_dh_missing_fail,
        ], rule_cb=override_dh_auto_test_drop_options)


upgrade_to_debhelper = {
    10: upgrade_to_debhelper_10,
    11: upgrade_to_debhelper_11,
    12: upgrade_to_debhelper_12,
    13: upgrade_to_debhelper_13,
}


for version in range(int(str(current_debhelper_compat_version))+1,
                     int(str(new_debhelper_compat_version))+1):
    with suppress(KeyError):
        upgrade_to_debhelper[version]()

if new_debhelper_compat_version > current_debhelper_compat_version:
    if current_debhelper_compat_version < lowest_non_deprecated_compat_level():
        kind = "deprecated"
        fixed_lintian_tag(
            'source', "package-uses-deprecated-debhelper-compat-version",
            info=f'{current_debhelper_compat_version}')
    else:
        kind = "old"
        fixed_lintian_tag(
            'source', "package-uses-old-debhelper-compat-version",
            info=f'{current_debhelper_compat_version}')
    description = "Bump debhelper from {} {} to {}.".format(
        kind, current_debhelper_compat_version, new_debhelper_compat_version)

    report_result(description=description, details=sorted(subitems))



#!/usr/bin/python3

from contextlib import suppress

from debmutate.reformatting import check_generated_file

from lintian_brush.fixer import LintianIssue, report_result
from lintian_brush.line_editor import LineEditor

check_generated_file('debian/rules')

with suppress(FileNotFoundError), LineEditor('debian/rules', 'b') as e:
    for lineno, line in e:
        newline = line.replace(
            b'$(dir $(_))', b'$(dir $(firstword $(MAKEFILE_LIST)))')
        if newline != line:
            issue = LintianIssue(
                'source', 'debian-rules-uses-special-shell-variable',
                info='line %d' % lineno)
            if issue.should_fix():
                e[lineno] = newline
                issue.report_fixed()


report_result('Avoid using $(_) to discover source package directory.')



#!/usr/bin/python3

from urllib.parse import urlparse

from lintian_brush.fixer import LintianIssue, control, report_result
from lintian_brush.lintian import is_obsolete_site

with control:
    homepage = control.source.get('Homepage')
    if homepage is not None and is_obsolete_site(urlparse(homepage)):
        issue = LintianIssue(
            control.source, 'obsolete-url-in-packaging', 'debian/control')
        if issue.should_fix():
            issue.report_fixed()
            del control.source['Homepage']


# TODO(jelmer): Check debian/copyright
# TODO(jelmer): Check debian/watch


report_result('Drop fields with obsolete URLs.')



#!/usr/bin/python3

import os
import re
import sys

from lintian_brush.fixer import LintianIssue, report_result
from lintian_brush.line_editor import LineEditor

if not os.path.isdir('debian/tests'):
    sys.exit(0)

for entry in os.scandir('debian/tests'):
    if not entry.is_file():
        continue
    with LineEditor(entry.path, 'b') as e:
        for lineno, oldline in e:
            newline = re.sub(b'\\bADTTMP\\b', b'AUTOPKGTEST_TMP', oldline)
            if oldline == newline:
                continue
            issue = LintianIssue(
                'source', 'uses-deprecated-adttmp',
                info='%s (line %d)' % (entry.path, lineno))
            if issue.should_fix():
                e[lineno] = newline
                issue.report_fixed()


report_result(
    'Replace use of deprecated $ADTTMP with $AUTOPKGTEST_TMP.',
    certainty='certain')



#!/usr/bin/python3
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Utility functions for dealing with changelog files."""

from datetime import datetime
from typing import List, Optional, Tuple

from breezy.mutabletree import MutableTree
from debmutate.changelog import (
    Changelog,
)
from debmutate.changelog import (
    changelog_add_entry as _changelog_add_entry,
)

DEFAULT_URGENCY = "medium"


def add_changelog_entry(
    tree: MutableTree,
    path: str,
    summary: List[str],
    maintainer: Optional[Tuple[str, str]] = None,
    timestamp: Optional[datetime] = None,
    urgency: str = DEFAULT_URGENCY,
) -> None:
    """Add a changelog entry.

    Args:
      tree: Tree to edit
      path: Path to the changelog file
      summary: Entry to add
      maintainer: Maintainer details; tuple of fullname and email
      suppress_warnings: Whether to suppress any warnings from 'dch'
    """
    # TODO(jelmer): This logic should ideally be in python-debian.
    with tree.get_file(path) as f:
        cl = Changelog()
        cl.parse_changelog(
                f, max_blocks=None, allow_empty_author=True,
                strict=False)
        _changelog_add_entry(
            cl,
            summary=summary,
            maintainer=maintainer,
            timestamp=timestamp,
            urgency=urgency,
        )
    # Workaround until
    # https://salsa.debian.org/python-debian-team/python-debian/-/merge_requests/22
    # lands.
    pieces = []
    for line in cl.initial_blank_lines:
        pieces.append(line.encode(cl._encoding) + b"\n")
    for block in cl._blocks:
        try:
            serialized = block._format(allow_missing_author=True).encode(
                block._encoding
            )
        except TypeError:  # older python-debian
            serialized = bytes(block)
        pieces.append(serialized)
    tree.put_file_bytes_non_atomic(path, b"".join(pieces))



#!/usr/bin/python
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Helper functions for fixers."""

import os
import sys
from collections.abc import MutableMapping
from typing import List, Optional, Tuple, Union

from debmutate.control import ControlEditor
from debmutate.deb822 import Deb822Paragraph
from debmutate.vendor import get_vendor_name

from debian.changelog import Version

from . import (
    DEFAULT_MINIMUM_CERTAINTY,
    certainty_sufficient,
)
from .lintian_overrides import (
    LintianOverride,
    get_overrides,
    load_renamed_tags,
)


class LintianIssue:
    """Represents a lintian issue."""

    target: Tuple[str, Optional[str]]
    info: Optional[str]
    tag: str

    def __init__(
        self,
        target: Union[Deb822Paragraph, MutableMapping, Tuple[str, str], str],
        tag: str,
        info: Optional[Union[str, Tuple[str, ...]]] = None,
    ):
        if isinstance(target, (Deb822Paragraph, MutableMapping)):
            if "Source" in target:
                self.target = ("source", target["Source"])
            elif "Package" in target:
                self.target = ("binary", target["Package"])
            else:
                raise ValueError(
                    "unable to determine source/binary package from target"
                )
        elif target == "source":
            self.target = ("source", None)
        else:
            self.target = target  # type: ignore
        if isinstance(info, tuple):
            self.info = " ".join(info)
        else:
            self.info = info
        self.tag = tag

    def json(self):
        return {
            "tag": self.tag,
            "info": self.info,
            "target": self.target,
        }

    def override_exists(self):
        return _override_exists(
            tag=self.tag, info=self.info, type=self.target[0],
            package=self.target[1]
        )

    def should_fix(self):
        if self.override_exists():
            _overriden_issues.append(self)
            return False
        return True

    def report_fixed(self):
        _fixed_lintian_issues.append(self)

    def __str__(self):
        ret = []
        if self.target[1] is not None:
            ret.append(self.target[1] + " ")
        ret.append(self.target[0])
        ret.append(": " + self.tag + (' ' + self.info) if self.info else '')
        return "".join(ret)

    def __repr__(self):
        return "{}(target={!r}, tag={!r}, info={!r})".format(
            type(self).__name__,
            self.target,
            self.tag,
            self.info,
        )


_fixed_lintian_issues: List[LintianIssue] = []
_present_overrides: Optional[List[LintianOverride]] = None
_overriden_issues: List[LintianIssue] = []
_tag_renames = None


def _override_exists(
    tag: str,
    info: Optional[str] = None,
    package: Optional[str] = None,
    type: Optional[str] = None,
    arch: Optional[str] = None,
) -> bool:
    global _present_overrides, _tag_renames
    if _present_overrides is None:
        _present_overrides = list(get_overrides())
    if not _present_overrides:
        return False
    if _tag_renames is None:
        _tag_renames = load_renamed_tags()
    for override in _present_overrides:
        if _tag_renames.get(override.tag) == tag:
            tag = override.tag
        if override.matches(
                package=package, info=info, tag=tag, arch=arch, type=type):
            return True
    return False


def fixed_lintian_tags():
    return {issue.tag for issue in _fixed_lintian_issues}


def fixed_lintian_tag(
    target: Union[MutableMapping, Tuple[str, str], str],
    tag: str,
    info: Optional[Union[str, Tuple[str, ...]]] = None,
):
    """Register a lintian tag as being fixed."""
    LintianIssue(target, tag, info).report_fixed()


def reset() -> None:
    """Reset any global state that may exist."""
    global _fixed_lintian_issues, _present_overrides, _overriden_issues
    _fixed_lintian_issues = []
    _present_overrides = None
    _overriden_issues = []


def report_result(description=None, certainty=None, patch_name=None,
                  details=None):
    """Report the result of a fixer.

    Args:
      description: Description of the fix
      certainty: Certainty of the fix
      patch_name: Suggested patch name, if there are upstream changes
      details: List of details, as lines
    """
    if description:
        print(description)
    if details:
        for detail in details:
            print('+ %s' % detail)
    if certainty:
        print("Certainty: %s" % certainty)
    fixed_lintian_tags = {
        issue.tag for issue in _fixed_lintian_issues}
    if fixed_lintian_tags:
        print("Fixed-Lintian-Tags: %s" % ", ".join(sorted(fixed_lintian_tags)))
    if _overriden_issues:
        print("Overridden-Lintian-Issues:")
        for issue in _overriden_issues:
            print(' ' + str(issue))
    if patch_name:
        print("Patch-Name: %s" % patch_name)
    reset()


def net_access_allowed():
    """Check whether network access is allowed."""
    return os.environ.get("NET_ACCESS", "disallow") == "allow"


def compat_release():
    """Codename of oldest release to stay compatible with."""
    return os.environ.get("COMPAT_RELEASE", "sid")


def upgrade_release():
    """Codename of oldest release to allow upgrading from."""
    return os.environ.get("UPGRADE_RELEASE", "oldstable")


def current_package_version():
    return Version(os.environ["CURRENT_VERSION"])


def package_is_native():
    return not current_package_version().debian_revision


def meets_minimum_certainty(certainty):
    return certainty_sufficient(
        certainty,
        os.environ.get("MINIMUM_CERTAINTY", DEFAULT_MINIMUM_CERTAINTY)
    )


def trust_package():
    return os.environ.get("TRUST_PACKAGE") == "true"


def opinionated():
    return os.environ.get("OPINIONATED", "no") == "yes"


def warn(msg):
    sys.stderr.write("%s\n" % msg)


def diligence():
    return int(os.environ.get("DILIGENCE", "0"))


def source_package_name():
    return os.environ.get("DEB_SOURCE")


def is_debcargo_package():
    return os.path.exists('debian/debcargo.toml')


if is_debcargo_package():
    from debmutate.debcargo import DebcargoControlShimEditor
    control = DebcargoControlShimEditor.from_debian_dir('debian')
else:
    control = ControlEditor()


def vendor() -> str:
    return get_vendor_name()


def linenos_to_ranges(linenos):
    ret = []
    cur_range: List[int] = []

    def finalize_range(r):
        if len(r) == 1:
            ret.append("%d" % r[0])
        else:
            ret.append("%d-%d" % (r[0], r[-1]))

    for lineno in linenos:
        if not cur_range or cur_range[-1] == lineno-1:
            cur_range.append(lineno)
        else:
            finalize_range(cur_range)
            cur_range = [lineno]
    if cur_range:
        finalize_range(cur_range)
    return ret


def shorten_path(path):
    ps = path.split('/')
    if ps[0] == 'debian':
        ps[0] = 'd'
    return '/'.join(ps)



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for lintian_brush.detect_gbp_dch."""

from breezy.tests import (
    TestCaseWithTransport,
)

from ..detect_gbp_dch import (
    ChangelogBehaviour,
    guess_update_changelog,
)


def make_changelog(entries):
    return (
        """\
lintian-brush (0.1) UNRELEASED; urgency=medium

"""
        + "".join(["  * %s\n" % entry for entry in entries])
        + """

 -- Jelmer Vernooij <jelmer@debian.org>  Sat, 13 Oct 2018 11:21:39 +0100
"""
    ).encode("utf-8")


class GuessUpdateChangelogTests(TestCaseWithTransport):
    def test_no_gbp_conf(self):
        tree = self.make_branch_and_tree(".")
        self.assertEqual(
            ChangelogBehaviour(
                True,
                'Assuming changelog needs to be updated, '
                'since it is always changed together '
                'with other files in the tree.'),
            guess_update_changelog(tree, "debian"))

    def test_custom_path(self):
        tree = self.make_branch_and_tree(".")
        self.assertEqual(ChangelogBehaviour(
            True,
            'Assuming changelog needs to be updated, '
            'since it is always changed together '
            'with other files in the tree.'),
            guess_update_changelog(tree, "debian"))
        self.assertEqual(ChangelogBehaviour(
            True,
            'assuming changelog needs to be updated since '
            'gbp dch only supports a debian '
            'directory in the root of the repository'),
            guess_update_changelog(tree, ""))
        self.assertEqual(ChangelogBehaviour(
            True,
            'assuming changelog needs to be updated since '
            'gbp dch only supports a debian '
            'directory in the root of the repository'),
            guess_update_changelog(tree, "lala/debian"))

    def test_gbp_conf_dch(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/gbp.conf",
                    """\
[dch]
pristine-tar = False
""",
                ),
            ]
        )
        tree.add(["debian", "debian/gbp.conf"])
        self.assertEqual(
            ChangelogBehaviour(
                False,
                "Assuming changelog does not need to be updated, since "
                "there is a [dch] section in gbp.conf.",
            ),
            guess_update_changelog(tree, "debian"),
        )

    def test_changelog_sha_prefixed(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/changelog",
                    """\
blah (0.20.1) unstable; urgency=medium

  [ Somebody ]
  * [ebb7c31] do a thing
  * [629746a] do another thing that actually requires us to wrap lines
    and then

  [ Somebody Else ]
  * [b02b435] do another thing

 -- Joe User <joe@example.com>  Tue, 19 Nov 2019 15:29:47 +0100
""",
                ),
            ]
        )
        tree.add(["debian", "debian/changelog"])
        self.assertEqual(
            ChangelogBehaviour(
                False,
                "Assuming changelog does not need to be updated, "
                "since all entries in last changelog entry are prefixed "
                "by git shas.",
            ),
            guess_update_changelog(tree, "debian"),
        )

    def test_empty(self):
        tree = self.make_branch_and_tree(".")
        self.assertEqual(
            ChangelogBehaviour(
                True,
                "Assuming changelog needs to be updated, "
                "since it is always changed together "
                "with other files in the tree.",
            ),
            guess_update_changelog(tree, "debian"),
        )

    def test_update_with_change(self):
        builder = self.make_branch_builder(".")
        builder.start_series()
        builder.build_snapshot(
            None,
            [
                ("add", ("", None, "directory", "")),
                ("add", ("upstream", None, "file", b"upstream")),
                ("add", ("debian/", None, "directory", "")),
                (
                    "add",
                    (
                        "debian/changelog",
                        None,
                        "file",
                        make_changelog(["initial release"]),
                    ),
                ),
                ("add", ("debian/control", None, "file", b"initial")),
            ],
            message="Initial\n",
        )
        changelog_entries = ["initial release"]
        for i in range(20):
            builder.build_snapshot(
                None, [("modify", ("upstream", b"upstream %d" % i))],
                message="Upstream"
            )
            changelog_entries.append("next entry %d" % i)
            builder.build_snapshot(
                None,
                [
                    ("modify", (
                        "debian/changelog",
                        make_changelog(changelog_entries))),
                    ("modify", ("debian/control", b"next %d" % i)),
                ],
                message="Next",
            )
        builder.finish_series()
        branch = builder.get_branch()
        tree = branch.controldir.create_workingtree()
        self.assertEqual(
            ChangelogBehaviour(
                True,
                "Assuming changelog needs to be updated, "
                "since it is always changed together "
                "with other files in the tree.",
            ),
            guess_update_changelog(tree, "debian"),
        )

    def test_changelog_updated_separately(self):
        builder = self.make_branch_builder(".")
        builder.start_series()
        builder.build_snapshot(
            None,
            [
                ("add", ("", None, "directory", "")),
                ("add", ("debian/", None, "directory", "")),
                (
                    "add",
                    (
                        "debian/changelog",
                        None,
                        "file",
                        make_changelog(["initial release"]),
                    ),
                ),
                ("add", ("debian/control", None, "file", b"initial")),
            ],
            message="Initial\n",
        )
        changelog_entries = ["initial release"]
        for i in range(20):
            changelog_entries.append("next entry %d" % i)
            builder.build_snapshot(
                None, [("modify", ("debian/control", b"next %d" % i))],
                message="Next\n"
            )
        builder.build_snapshot(
            None,
            [("modify",
                ("debian/changelog", make_changelog(changelog_entries)))]
        )
        changelog_entries.append("final entry")
        builder.build_snapshot(
            None, [("modify", ("debian/control", b"more"))], message="Next\n"
        )
        builder.build_snapshot(
            None,
            [("modify",
                ("debian/changelog", make_changelog(changelog_entries)))]
        )
        builder.finish_series()
        branch = builder.get_branch()
        tree = branch.controldir.create_workingtree()
        self.assertEqual(
            ChangelogBehaviour(
                False,
                "Assuming changelog does not need to be updated, "
                "since changelog entries are usually updated in "
                "separate commits.",
            ),
            guess_update_changelog(tree, "debian"),
        )

    def test_has_dch_in_messages(self):
        builder = self.make_branch_builder(".")
        builder.build_snapshot(
            None,
            [("add", ("", None, "directory", ""))],
            message="Git-Dch: ignore\n"
        )
        branch = builder.get_branch()
        tree = branch.controldir.create_workingtree()
        self.assertEqual(
            ChangelogBehaviour(
                False,
                "Assuming changelog does not need to be updated, "
                "since there are Gbp-Dch stanzas in commit messages",
            ),
            guess_update_changelog(tree, "debian"),
        )

    def test_inaugural_unreleased(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/changelog",
                    """\
blah (0.20.1) UNRELEASED; urgency=medium

  * Initial release. Closes: #123123

 -- Joe User <joe@example.com>  Tue, 19 Nov 2019 15:29:47 +0100
""",
                ),
            ]
        )
        tree.add(["debian", "debian/changelog"])
        self.assertEqual(
            ChangelogBehaviour(
                False,
                'assuming changelog does not need to be '
                'updated since it is the inaugural unreleased entry'
            ),
            guess_update_changelog(tree, "debian"),
        )

    def test_last_entry_warns_generated(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/changelog",
                    """\
blah (0.20.1) UNRELEASED; urgency=medium

  * WIP (generated at release time: please do not add entries below).

 -- Joe User <joe@example.com>  Tue, 19 Nov 2019 15:29:47 +0100

blah (0.20.1) unstable; urgency=medium

  * Initial release. Closes: #123123

 -- Joe User <joe@example.com>  Tue, 19 Nov 2019 15:29:47 +0100
""",
                ),
            ]
        )
        tree.add(["debian", "debian/changelog"])
        self.assertEqual(
            ChangelogBehaviour(
                False,
                'last changelog entry warns changelog is generated '
                'at release time',
            ),
            guess_update_changelog(tree, "debian"),
        )

    def test_never_unreleased(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents(
            [
                ("debian/",),
                ("debian/control", "foo"),
                (
                    "debian/changelog",
                    """\
blah (0.20.1) unstable; urgency=medium

  * Initial release. Closes: #123123

 -- Joe User <joe@example.com>  Tue, 19 Nov 2019 15:29:47 +0100
""",
                ),
            ]
        )
        tree.add(["debian", "debian/control", "debian/changelog"])
        tree.commit("rev1")
        self.build_tree_contents([("debian/control", "bar")])
        tree.commit("rev2")
        self.build_tree_contents([("debian/control", "bla")])
        tree.commit("rev2")
        self.build_tree_contents(
            [
                (
                    "debian/changelog",
                    """\
blah (0.21.1) unstable; urgency=medium

  * Next release.

 -- Joe User <joe@example.com>  Tue, 19 Nov 2019 15:29:47 +0100

blah (0.20.1) unstable; urgency=medium

  * Initial release. Closes: #123123

 -- Joe User <joe@example.com>  Tue, 19 Nov 2019 15:29:47 +0100
""",
                ),
            ]
        )
        tree.commit("rev2")
        self.assertEqual(
            ChangelogBehaviour(
                False,
                'Assuming changelog does not need to be updated, '
                'since it never uses UNRELEASED entries'
            ),
            guess_update_changelog(tree, "debian"),
        )



#!/usr/bin/python3

import sys
from contextlib import suppress

from lintian_brush.fixer import report_result, warn
from lintian_brush.yaml import YamlUpdater

try:
    from Levenshtein import distance
except ImportError:
    sys.exit(2)

valid_field_names = {
    'Archive',
    'ASCL-Id',
    'Bug-Database',
    'Bug-Submit',
    'Cite-As',
    'Changelog',
    'CPE',
    'Documentation',
    'Donation',
    'FAQ',
    'Funding',
    'Gallery',
    'Other-References',
    'Reference',
    'Registration',
    'Registry',
    'Repository',
    'Repository-Browse',
    'Screenshots',
    'Security-Contact',
    'Webservice',
}

typo_fixed = set()
case_fixed = set()

with suppress(FileNotFoundError), \
        YamlUpdater('debian/upstream/metadata') as updater:
    for field in updater.code:
        if field in valid_field_names:
            continue
        if (field.startswith('X-') and
                field[2:] in valid_field_names):
            if field[2:] in updater.code:
                warn(f'Both {field} and {field[2:]} exist.')
                continue
            value = updater.code[field]
            del updater.code[field]
            updater.code[field[2:]] = value
            typo_fixed.add((field, field[2:]))
            continue

        for option in valid_field_names:
            if distance(field, option) == 1:
                value = updater.code[field]
                del updater.code[field]
                updater.code[option] = value
                if option.lower() == field.lower():
                    case_fixed.add((field, option))
                else:
                    typo_fixed.add((field, option))
                break


if case_fixed:
    kind = 'case' + ('s' if len(case_fixed) > 1 else '')
else:
    kind = ''
if typo_fixed:
    if case_fixed:
        kind += ' and '
    kind += 'typo' + ('s' if len(typo_fixed) > 1 else '')

fixed_str = ', '.join(
    [f'{old} ⇒ {new}'
     for (old, new) in sorted(list(case_fixed) + list(typo_fixed))])

report_result(
    f'Fix field name {kind} in debian/upstream/metadata ({fixed_str}).')



#!/usr/bin/python3

from lintian_brush.fixer import LintianIssue, control, report_result

with control as editor:
    for binary in control.binaries:
        try:
            desc = binary['Description']
        except KeyError:
            # Uhm, okay.
            continue
        lines = desc.splitlines(True)
        issue = LintianIssue(
            binary, 'extended-description-contains-empty-paragraph', ())
        if not issue.should_fix():
            continue
        if len(lines) > 1 and lines[1] == ' .\n':
            del lines[1]
        else:
            continue
        binary['Description'] = ''.join(lines)
        issue.report_fixed()

report_result('Remove empty leading paragraph in Description.')



#!/usr/bin/python3

import os
import sys

from lintian_brush.fixer import (
    LintianIssue,
    opinionated,
    report_result,
)
from lintian_brush.patches import read_quilt_series

if not opinionated():
    # In a lot of cases, it seems like removing the patch is not the right
    # thing to do.
    sys.exit(0)

try:
    patches = set()
    with open('debian/patches/series', 'rb') as f:
        for patch, _quoted, _options in read_quilt_series(f):
            patches.add(patch)
except FileNotFoundError:
    sys.exit(0)

removed = set()

for patch in os.listdir('debian/patches'):
    path = os.path.join('debian', 'patches', patch)
    # Don't delete the series file
    if patch in ('series', '00list'):
        continue
    # Ignore everything that is not a regular file
    if not os.path.isfile(path):
        continue
    # Ignore any README files
    if patch.startswith('README'):
        continue
    # Ignore any notes files
    if patch.startswith('notes'):
        continue
    # Ignore everything that is listed in series
    if patch in patches:
        continue
    issue = LintianIssue(
        'source', 'patch-file-present-but-not-mentioned-in-series',
        patch)
    if not issue.should_fix():
        continue
    removed.add(patch)
    os.unlink(path)
    issue.report_fixed()


description = (
      "Remove patch{} {} that {} missing from debian/patches/series.".format(
          'es' if len(removed) > 1 else '', ', '.join(sorted(removed)),
          'is' if len(removed) == 1 else 'are'))
report_result(description)



#!/usr/bin/python3

from email.utils import parseaddr
from typing import Dict, List, Optional, Tuple

from lintian_brush.fixer import control, fixed_lintian_tag, report_result

REPLACEMENTS: Dict[Optional[str], Tuple[str, List[Tuple[str, str]]]] = {
    'python-modules-team@lists.alioth.debian.org':
        ('old-dpmt-vcs',
         [('https://salsa.debian.org/python-team/modules/',
           'https://salsa.debian.org/python-team/packages/')]),
    'python-apps-team@lists.alioth.debian.org':
        ('old-papt-vcs',
         [('https://salsa.debian.org/python-team/applications/',
           'https://salsa.debian.org/python-team/packages/')]),
}

with control as editor:
    maint: Optional[str]
    email: Optional[str]
    try:
        maint, email = parseaddr(editor.source['Maintainer'])
    except KeyError:
        maint, email = None, None
    changed_fields = set()
    try:
        tag, replacements = REPLACEMENTS[email]
    except KeyError:
        pass
    else:
        for field in [f for f in editor.source if f.startswith('Vcs-')]:
            url = editor.source[field]
            for old, new in replacements:
                url = url.replace(old, new)
            if url != editor.source[field]:
                editor.source[field] = url
                changed_fields.add(field)
                fixed_lintian_tag(editor.source, tag, info='')

    report_result('Update fields {} for maintainer {}.'.format(
                  ', '.join(sorted(changed_fields)), maint))



#!/usr/bin/python3

import sys

from debian.changelog import get_maintainer
from lintian_brush.fixer import (
    control,
    fixed_lintian_tag,
    meets_minimum_certainty,
    report_result,
)

# TODO(jelmer): Bump this up if there's a way that we can verify that e.g. the
# ITP was filed by get_maintainer() ?
CERTAINTY = 'possible'

if not meets_minimum_certainty(CERTAINTY):
    sys.exit(0)


with control as updater:
    if updater.source.get('Maintainer'):
        sys.exit(0)
    maintainer = get_maintainer()
    updater.source['Maintainer'] = "{} <{}>".format(*maintainer)
    fixed_lintian_tag(updater.source, 'required-field', 'Maintainer')

report_result(
    'Set the maintainer field to: {} <{}>.'.format(*maintainer),
    certainty=CERTAINTY)



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij <jelmer@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from . import _lintian_brush_rs

report_fatal = _lintian_brush_rs.report_fatal
report_success = _lintian_brush_rs.report_success
report_success_debian = _lintian_brush_rs.report_success_debian
svp_enabled = _lintian_brush_rs.svp_enabled
load_resume = _lintian_brush_rs.load_resume



#!/usr/bin/python3
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Handling of quilt patches."""

__all__ = [
    "PatchSyntax",
    "find_patch_base",
    "find_patches_branch",
    "AppliedPatches",
    "read_quilt_patches",
    "PatchApplicationBaseNotFound",
    "upstream_with_applied_patches",
    "find_patches_directory",
    "rules_find_patches_directory",
    "tree_patches_directory",
    "add_patch",
    "move_upstream_changes_to_patch",
]

import contextlib
import os
from datetime import datetime
from email.message import Message
from io import BytesIO
from typing import List, Optional, Tuple

import breezy.bzr  # noqa: F401
import breezy.git  # noqa: F401
from breezy import osutils
from breezy.commit import filter_excluded
from breezy.diff import show_diff_trees
from breezy.errors import NotBranchError
from breezy.patches import (
    PatchSyntax,
    apply_patches,
    parse_patches,
)
from breezy.transport import NoSuchFile
from breezy.tree import Tree
from breezy.workingtree import WorkingTree
from breezy.workspace import reset_tree
from debmutate.patch import (
    QuiltSeriesEditor,
    find_common_patch_suffix,
    read_quilt_series,
)

from debian.changelog import Changelog

# TODO(jelmer): Use debmutate version
DEFAULT_DEBIAN_PATCHES_DIR = "debian/patches"


def find_patch_base(tree):
    """Find the base revision to apply patches to.

    Args:
      tree: Tree to find the patch base for
    Returns:
      A revision string
    """
    with tree.get_file("debian/changelog") as f:
        cl = Changelog(f, max_blocks=1)
        package = cl.package
        upstream_version = cl.version.upstream_version
    possible_tags = [
        "upstream-%s" % upstream_version,
        "upstream/%s" % upstream_version,
        "%s" % upstream_version,
        "v%s" % upstream_version,
        f"{package}-{upstream_version}",
    ]
    tags = tree.branch.tags.get_tag_dict()
    for possible_tag in possible_tags:
        if possible_tag in tags:
            return tags[possible_tag]
    # TODO(jelmer): Do something clever, like look for the last merge?
    return None


def find_patches_branch(tree):
    """Find the branch that is used to track patches.

    Args:
      tree: Tree for which to find patches branch
    Returns:
      A `Branch` instance
    """
    if tree.branch.name is None:
        return None
    branch_name = "patch-queue/%s" % tree.branch.name
    with contextlib.suppress(NotBranchError):
        return tree.branch.controldir.open_branch(branch_name)
    if tree.branch.name == "master":
        branch_name = "patched"
    else:
        branch_name = "patched-%s" % tree.branch.name
    with contextlib.suppress(NotBranchError):
        return tree.branch.controldir.open_branch(branch_name)
    return None


class AppliedPatches:
    """Context that provides access to a tree with patches applied."""

    def __init__(self, tree, patches, prefix=1):
        self.tree = tree
        self.patches = patches
        self.prefix = prefix

    def __enter__(self):
        if self.patches:
            self._tt = self.tree.preview_transform()
            apply_patches(self._tt, self.patches, prefix=self.prefix)
            return self._tt.get_preview_tree()
        else:
            self._tt = None
            return self.tree

    def __exit__(self, exc_type, exc_value, exc_tb):
        if self._tt:
            self._tt.finalize()
        return False


def read_quilt_patches(tree, directory=DEFAULT_DEBIAN_PATCHES_DIR):
    """Read patch contents from quilt directory.

    Args:
      tree: Tree to read
      directory: Patch directory
    Returns:
      list of Patch objects
    """
    series_path = osutils.pathjoin(directory, "series")
    try:
        series_lines = tree.get_file_lines(series_path)
    except NoSuchFile:
        return []
    for entry in read_quilt_series(series_lines):
        if entry.quoted:
            continue
        # TODO(jelmer): Pass on options?
        with tree.get_file(osutils.pathjoin(directory, entry.name)) as f:
            yield from parse_patches(f, allow_dirty=True, keep_dirty=False)


class PatchApplicationBaseNotFound(Exception):
    """Unable to find base for patch application."""


@contextlib.contextmanager
def upstream_with_applied_patches(tree, patches):
    """Create a copy of the upstream tree with applied patches.

    Args:
      tree: Tree with applied patches
    """
    patches_branch = find_patches_branch(tree)
    if patches_branch is not None:
        # TODO(jelmer): Make sure it's actually rebased on current upstream
        yield patches_branch.basis_tree()
        return

    upstream_revision = find_patch_base(tree)
    if upstream_revision is None:
        raise PatchApplicationBaseNotFound(tree)
    upstream_tree = tree.branch.repository.revision_tree(upstream_revision)

    with AppliedPatches(upstream_tree, patches) as tree:
        yield tree


def rules_find_patches_directory(makefile):
    """Find the patches directory set in debian/rules.

    Args:
        makefile: Makefile to scan
    Returns:
        path to patches directory, or None if none was found in debian/rules
    """
    try:
        val = makefile.get_variable(b"QUILT_PATCH_DIR")
    except KeyError:
        return None
    else:
        return val.decode()


def find_patches_directory(path):
    """Find the name of the patches directory, if any.

    Args:
      path: Root to package
    Returns:
      relative path to patches directory, or None if none exists
    """
    from debmutate._rules import Makefile

    directory = None
    try:
        mf = Makefile.from_path(os.path.join(path, "debian/rules"))
    except FileNotFoundError:
        pass
    else:
        rules_directory = rules_find_patches_directory(mf)
        if rules_directory is not None:
            directory = rules_directory
    if directory is None and os.path.exists(
        os.path.join(path, DEFAULT_DEBIAN_PATCHES_DIR)
    ):
        directory = DEFAULT_DEBIAN_PATCHES_DIR
    return directory


def tree_patches_directory(tree, subpath=""):
    """Find the name of the patches directory.

    This will always return a path, even if the patches
    directory does not yet exist.

    Args:
      tree: Tree to check
    Returns:
      path to patches directory, or what it should be
    """
    directory = find_patches_directory(tree.abspath(subpath))
    if directory is None:
        return DEFAULT_DEBIAN_PATCHES_DIR
    return directory


def tree_non_patches_changes(tree, patches_directory):
    """Check if a Debian tree has changes vs upstream tree.

    Args:
      tree: Tree to check
      patches_directory: Name of patches directory
    Returns:
        list of TreeDelta objects
    """
    if patches_directory is None:
        patches = []
    else:
        patches = list(read_quilt_patches(tree, patches_directory))

    # TODO(jelmer): What if patches are already applied on tree?
    with upstream_with_applied_patches(
        tree, patches
    ) as upstream_patches_tree, AppliedPatches(tree, patches) as patches_tree:
        for change in filter_excluded(
            patches_tree.iter_changes(
                upstream_patches_tree), exclude=["debian"]
        ):
            path = change.path[1]
            if path == "":
                continue
            yield change


def add_patch(tree, patches_directory, name, contents, header=None):
    """Add a new patch.

    Args:
      tree: Tree to edit
      patches_directory: Name of patches directory
      name: Patch name without suffix
      contents: Diff
      header: RFC822 to read
    Returns:
      Name of the patch that was written (including suffix)
    """
    if not tree.has_filename(patches_directory):
        if not tree.has_filename(os.path.dirname(patches_directory)):
            tree.mkdir(os.path.dirname(patches_directory))
        tree.mkdir(patches_directory)
    abs_patches_dir = tree.abspath(patches_directory)
    patch_suffix = find_common_patch_suffix(os.listdir(abs_patches_dir))
    patchname = name + patch_suffix
    path = os.path.join(abs_patches_dir, patchname)
    if tree.has_filename(path):
        raise FileExistsError(path)
    with open(tree.abspath(path), "wb") as f:
        if header is not None:
            f.write(header.as_string().encode("utf-8"))
        f.write(contents)

    # TODO(jelmer): Write to patches branch if applicable

    series_path = os.path.join(abs_patches_dir, "series")
    with QuiltSeriesEditor(series_path) as editor:
        editor.append(patchname)

    return patchname


def move_upstream_changes_to_patch(
    local_tree: WorkingTree,
    basis_tree: Tree,
    subpath: str,
    patch_name: str,
    description: str,
    dirty_tracker=None,
    timestamp: Optional[datetime] = None,
) -> Tuple[List[str], str]:
    """Move upstream changes to patch.

    Args:
      local_tree: local tree
      subpath: subpath
      patch_name: Suggested patch name
      description: Description
      dirty_tracker: Dirty tracker
    """
    if timestamp is None:
        timestamp = datetime.now()
    diff = BytesIO()
    show_diff_trees(basis_tree, local_tree, diff)
    reset_tree(local_tree, basis_tree, subpath, dirty_tracker)
    header = Message()
    lines = description.splitlines()
    # See https://dep-team.pages.debian.net/deps/dep3/ for fields.
    header["Description"] = (
        lines[0]
        + "\n"
        + "\n".join([(" " + line) if line else " ." for line in lines[1:]])
    )
    header["Origin"] = "other"
    header["Last-Update"] = timestamp.strftime("%Y-%m-%d")
    patches_directory = tree_patches_directory(local_tree, subpath)
    patchname = add_patch(
        local_tree,
        os.path.join(subpath, patches_directory),
        patch_name,
        diff.getvalue(),
        header,
    )
    specific_files = [
        os.path.join(subpath, patches_directory),
        os.path.join(subpath, patches_directory, "series"),
        os.path.join(subpath, patches_directory, patchname),
    ]
    local_tree.add(specific_files)
    return specific_files, patchname



#!/usr/bin/python3

import os
import sys
from typing import Optional

from debmutate.control import ensure_some_version

from lintian_brush.fixer import (
    control,
    fixed_lintian_tag,
    is_debcargo_package,
    report_result,
)

if is_debcargo_package():
    sys.exit(0)

format: Optional[str]
try:
    with open('debian/source/format') as f:
        format = f.read().strip()
except FileNotFoundError:
    format = None

if format != '3.0 (quilt)' and os.path.exists('debian/patches/series'):
    with control as updater:
        updater.source['Build-Depends'] = ensure_some_version(
            updater.source.get('Build-Depends', ''), 'quilt')
    if updater.changed:
        fixed_lintian_tag(updater.source, 'quilt-series-but-no-build-dep')


report_result('Add missing dependency on quilt.')



#!/usr/bin/python3

import sys
from email.utils import parseaddr

# Import convenience functions for reporting results and checking overrides
from lintian_brush.fixer import LintianIssue, control, report_result

PKG_PERL_EMAIL = 'pkg-perl-maintainers@lists.alioth.debian.org'
TESTSUITE_VALUE = 'autopkgtest-pkg-perl'


with control as e:
    # Parse the maintainer field and extract the email address.
    try:
        (name, email) = parseaddr(e.source['Maintainer'])
    except KeyError:
        sys.exit(2)
    if email != PKG_PERL_EMAIL:
        # Nothing to do here, it's not a pkg-perl-maintained package
        sys.exit(0)
    if e.source.get('Testsuite') == TESTSUITE_VALUE:
        sys.exit(0)
    issue = LintianIssue(
        e.source, 'team/pkg-perl/testsuite/no-testsuite-header',
        info="autopkgtest")
    if issue.should_fix():
        e.source['Testsuite'] = TESTSUITE_VALUE
        issue.report_fixed()

report_result(
    'Set Testsuite header for perl package.',
    certainty='certain')



#!/usr/bin/python3

from typing import Optional

from lintian_brush.fixer import LintianIssue, control, report_result

packages = []


with control as updater:
    default_priority = updater.source.get("Priority")

    # TODO(jelmer): needs higher certainty?
    for binary in updater.binaries:
        if binary.get("Package-Type") == "udeb":
            continue
        if "transitional package" not in binary.get("Description", ""):
            continue
        oldsection = binary.get('Section') or updater.source.get('Section')
        issue = LintianIssue(
            binary, 'transitional-package-not-oldlibs-optional',
            '{}/{}'.format(
                oldsection,
                binary.get('Priority') or updater.source.get('Priority')))
        if issue.should_fix():
            packages.append(binary["Package"])
            area: Optional[str]
            if oldsection and '/' in oldsection:
                area, oldsection = oldsection.split('/', 1)
            else:
                area = None
            if area:
                binary["Section"] = f"{area}/oldlibs"
            else:
                binary["Section"] = "oldlibs"
            if default_priority != "optional":
                binary["Priority"] = "optional"
            elif "Priority" in binary:
                del binary["Priority"]
            issue.report_fixed()


report_result(
    "Move transitional package{} {} to oldlibs/optional per policy 4.0.1."
    .format(("s" if len(packages) > 1 else ""), ", ".join(packages)))



#!/usr/bin/python3

import sys
from contextlib import suppress

from debmutate.deb822 import Deb822Editor

from lintian_brush.fixer import fixed_lintian_tag, report_result, warn

try:
    from Levenshtein import distance
except ImportError:
    sys.exit(2)

valid_field_names = {
    'Files', 'License', 'Copyright', 'Comment',
    'Upstream-Name', 'Format', 'Upstream-Contact',
    'Source', 'Upstream', 'Contact', 'Name'}

typo_fixed = set()
case_fixed = set()

with suppress(FileNotFoundError), Deb822Editor('debian/copyright') as updater:
    for paragraph in updater.paragraphs:
        for field in paragraph:
            if field in valid_field_names:
                continue
            if (field.startswith('X-') and
                    field[2:] in valid_field_names):
                if field[2:] in paragraph:
                    warn(f'Both {field} and {field[2:]} exist.')
                    continue
                value = paragraph[field]
                del paragraph[field]
                paragraph[field[2:]] = value
                typo_fixed.add((field, field[2:]))
                fixed_lintian_tag(
                    'source', 'field-name-typo-in-dep5-copyright',
                    '%s (line XX)' % field)
                continue

            for option in valid_field_names:
                if distance(field, option) == 1:
                    value = paragraph[field]
                    if (option in paragraph
                            and option.lower() != field.lower()):
                        warn(
                            f'Found typo ({field} ⇒ {option}), '
                            f'but {option} already exists')
                        continue
                    del paragraph[field]
                    paragraph[option] = value
                    if option.lower() == field.lower():
                        case_fixed.add((field, option))
                    else:
                        typo_fixed.add((field, option))
                        fixed_lintian_tag(
                            'source', 'field-name-typo-in-dep5-copyright',
                            '%s (line XX)' % field)
                    break


if case_fixed:
    kind = 'case' + ('s' if len(case_fixed) > 1 else '')
else:
    kind = ''
if typo_fixed:
    if case_fixed:
        kind += ' and '
    kind += 'typo' + ('s' if len(typo_fixed) > 1 else '')

fixed_str = ', '.join(
    [f'{old} ⇒ {new}'
     for (old, new) in sorted(list(case_fixed) + list(typo_fixed))])

report_result(
    f'Fix field name {kind} in debian/copyright ({fixed_str}).')



#!/usr/bin/python3

import json
import sys
from urllib.request import urlopen

BASE_URL = (
    'https://raw.githubusercontent.com/spdx/license-list-data/master/json/')
LICENSES_URL = BASE_URL + 'licenses.json'
EXCEPTIONS_URL = BASE_URL + 'exceptions.json'

licenses_summary = json.loads(urlopen(LICENSES_URL).read())
licenses = {}
license_ids = []
for license in licenses_summary['licenses']:
    license_ids.append(license['licenseId'])
    licenses[license['licenseId']] = {
        'name': license['name'],
        }

exceptions = json.loads(urlopen(EXCEPTIONS_URL).read())
exception_ids = []
for exception in exceptions['exceptions']:
    exception_ids.append(exception['licenseExceptionId'])
result = {'licenses': licenses, 'exception_ids': exception_ids}
json.dump(result, sys.stdout, sort_keys=True, indent=4)



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij <jelmer@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

# See https://wiki.debian.org/UpstreamMetadata and
# https://dep-team.pages.debian.net/deps/dep12/

ADDON_ONLY_FIELDS = ["Archive"]
DEP12_FIELDS = [
    'Name',
    'Contact',
    'Archive',
    'ASCL-Id',
    'Bug-Database',
    'Bug-Submit',
    'Cite-As',
    'Changelog',
    'CPE',
    'Documentation',
    'Donation',
    'FAQ',
    'Funding',
    'Gallery',
    'Other-References',
    'Reference',
    'Registration',
    'Registry',
    'Repository',
    'Repository-Browse',
    'Screenshots',
    'Security-Contact',
    'Webservice',
]



#!/usr/bin/python3

from debmutate._rules import (
    dh_invoke_drop_with,
    update_rules,
)
from debmutate.debhelper import (
    ensure_minimum_debhelper_version,
)

from lintian_brush.fixer import (
    LintianIssue,
    control,
    report_result,
)


def cb(line, target):
    if line.strip() == b'dh_autotools-dev_updateconfig':
        issue = LintianIssue(
            'source', 'debhelper-tools-from-autotools-dev-are-deprecated',
            info='dh_autotools-dev_updateconfig')
        if issue.should_fix():
            issue.report_fixed()
            return []
    if line.strip() == b'dh_autotools-dev_restoreconfig':
        issue = LintianIssue(
            'source', 'debhelper-tools-from-autotools-dev-are-deprecated',
            info='dh_autotools-dev_restoreconfig')
        if issue.should_fix():
            issue.report_fixed()
            return []
    newline = dh_invoke_drop_with(line, b'autotools-dev')
    if newline != line:
        issue = LintianIssue(
            'source', 'debhelper-tools-from-autotools-dev-are-deprecated',
            info='dh ... --with autotools-dev')
        if issue.should_fix():
            line = newline
            issue.report_fixed()
    newline = dh_invoke_drop_with(line, b'autotools_dev')
    if newline != line:
        issue = LintianIssue(
            'source', 'debhelper-tools-from-autotools-dev-are-deprecated',
            info='dh ... --with autotools_dev')
        if issue.should_fix():
            line = newline
            issue.report_fixed()
    return line


if update_rules(cb):
    with control as updater:
        ensure_minimum_debhelper_version(updater.source, "9.20160114")


report_result("Drop use of autotools-dev debhelper.")



#!/usr/bin/python3

import re
import socket
import urllib.error
from contextlib import suppress
from email.utils import parseaddr
from urllib.parse import urlparse
from urllib.request import Request, urlopen

from debmutate.vcs import (
    get_vcs_info,
)

from lintian_brush import DEFAULT_URLLIB_TIMEOUT, USER_AGENT
from lintian_brush.fixer import (
    control,
    fixed_lintian_tag,
    net_access_allowed,
    report_result,
    warn,
)
from lintian_brush.salsa import (
    determine_browser_url as determine_salsa_browser_url,
)
from lintian_brush.salsa import (
    guess_repository_url,
    salsa_url_from_alioth_url,
)
from lintian_brush.vcs import (
    determine_browser_url,
)
from lintian_brush.vcswatch import VcsWatch, VcsWatchError

OBSOLETE_HOSTS = [
    'anonscm.debian.org', 'alioth.debian.org', 'svn.debian.org',
    'git.debian.org', 'bzr.debian.org', 'hg.debian.org']


def is_on_obsolete_host(url):
    parsed_url = urlparse(url)
    host = parsed_url.netloc.split('@')[-1]
    return host in OBSOLETE_HOSTS


def verify_salsa_repository(url):
    headers = {'User-Agent': USER_AGENT}
    browser_url = determine_salsa_browser_url(url)
    try:
        response = urlopen(
            Request(browser_url, headers=headers),
            timeout=DEFAULT_URLLIB_TIMEOUT)
    except socket.timeout:
        return None
    return response.status == 200


def retrieve_vcswatch_urls(package):
    try:
        with VcsWatch() as vcs_watch:
            return vcs_watch.get_package(package)
    except ImportError as exc:
        # No psycopg2, nothing
        raise KeyError(package) from exc


class NewRepositoryURLUnknown(Exception):

    def __init__(self, vcs_type, vcs_url):
        self.vcs_type = vcs_type
        self.vcs_url = vcs_url


def find_new_urls(vcs_type, vcs_url, package, maintainer_email,
                  net_access=False):
    if net_access and (
            vcs_url.startswith('https://') or vcs_url.startswith('http://')):
        headers = {'User-Agent': USER_AGENT}
        try:
            response = urlopen(
                Request(vcs_url, headers=headers),
                timeout=DEFAULT_URLLIB_TIMEOUT)
        except (urllib.error.HTTPError, urllib.error.URLError):
            pass
        except socket.timeout:
            pass
        else:
            redirected_url = response.geturl()
            if not is_on_obsolete_host(redirected_url):
                vcs_url = redirected_url
                vcs_browser = determine_browser_url(vcs_type, vcs_url)
                print("Update Vcs-* headers from URL redirect.")
                return (vcs_type, vcs_url, vcs_browser)

    # If possible, we use vcswatch to find the VCS repository URL
    if net_access:
        try:
            (vcs_type, vcs_url, vcs_browser) = retrieve_vcswatch_urls(package)
        except VcsWatchError as e:
            warn('vcswatch URL unusable: %s' % e.args[0])
        except KeyError:
            pass
        else:
            if not is_on_obsolete_host(vcs_url):
                print("Update Vcs-* headers from vcswatch.")
                if is_on_obsolete_host(vcs_browser):
                    vcs_browser = (
                        determine_browser_url(vcs_type, vcs_url) or
                        vcs_browser)
                return (vcs_type, vcs_url, vcs_browser)
            warn('vcswatch URL %s is still on old infrastructure.' % vcs_url)

    # Otherwise, attempt to guess based on maintainer email.
    guessed_url = guess_repository_url(package, maintainer_email)
    if guessed_url is not None:
        vcs_type = "Git"
        vcs_url = guessed_url
    else:
        vcs_url = salsa_url_from_alioth_url(vcs_type, vcs_url)
        if vcs_url is None:
            raise NewRepositoryURLUnknown(vcs_type, vcs_url)
        vcs_type = "Git"
    # Verify that there is actually a repository there
    if net_access and verify_salsa_repository(vcs_url) is False:
        raise NewRepositoryURLUnknown(vcs_type, vcs_url)

    print("Update Vcs-* headers to use salsa repository.")

    vcs_browser = determine_salsa_browser_url(vcs_url)
    return (vcs_type, vcs_url, vcs_browser)


def migrate_from_obsolete_infra(control):
    vcs_type, vcs_url, unused_subpath = get_vcs_info(control)
    if vcs_type is None:
        return
    if not is_on_obsolete_host(vcs_url):
        return

    package = control["Source"]
    maintainer_email = parseaddr(control["Maintainer"])[1]

    old_vcs_browser = control.get('Vcs-Browser')
    old_vcs_type = vcs_type
    old_vcs_url = vcs_url
    try:
        (vcs_type, vcs_url, vcs_browser) = find_new_urls(
            vcs_type, vcs_url, package, maintainer_email,
            net_access=net_access_allowed())
    except NewRepositoryURLUnknown:
        return

    fixed_lintian_tag(
        'source', "vcs-obsolete-in-debian-infrastructure",
        info=f'vcs-{old_vcs_type.lower()} {old_vcs_url}')

    if (("Vcs-Cvs" in control and re.match(
            r"\@(?:cvs\.alioth|anonscm)\.debian\.org:/cvsroot/",
            control["Vcs-Cvs"])) or
        ("Vcs-Svn" in control and
            "viewvc" in control["Vcs-Browser"])):
        fixed_lintian_tag(
            'source', "vcs-field-bitrotted",
            info='{} {}'.format(old_vcs_url or '', old_vcs_browser or ''))

    for hdr in ["Vcs-Git", "Vcs-Bzr", "Vcs-Hg", "Vcs-Svn"]:
        if hdr == "Vcs-" + vcs_type:  # type: ignore
            continue
        with suppress(KeyError):
            del control[hdr]
    control["Vcs-" + vcs_type] = vcs_url  # type: ignore
    if vcs_browser is not None:
        control["Vcs-Browser"] = vcs_browser
    else:
        with suppress(KeyError):
            del control["Vcs-Browser"]


with control as updater:
    migrate_from_obsolete_infra(updater.source)

report_result()



#!/usr/bin/python3

import os
import sys

from debmutate.control import delete_from_list
from debmutate.deb822 import Deb822Editor

from lintian_brush.fixer import fixed_lintian_tag, report_result
from lintian_brush.lintian import LINTIAN_DATA_PATH

removed_restrictions = []


if not os.path.exists('debian/tests/control'):
    sys.exit(0)

KNOWN_OBSOLETE_RESTRICTIONS_PATH = (
    os.path.join(LINTIAN_DATA_PATH, 'testsuite/known-obsolete-restrictions'))
DEPRECATED_RESTRICTIONS = []

try:
    with open(KNOWN_OBSOLETE_RESTRICTIONS_PATH) as f:
        for line in f:
            if line.startswith('#'):
                continue
            if not line.strip():
                continue
            DEPRECATED_RESTRICTIONS.append(line.strip())
except FileNotFoundError:
    sys.exit(2)

with Deb822Editor('debian/tests/control') as updater:
    for paragraph in updater.paragraphs:
        restrictions = paragraph.get('Restrictions', '').split(',')
        if restrictions == ['']:
            continue
        to_delete = []
        for restriction in list(restrictions):
            if restriction.strip() in DEPRECATED_RESTRICTIONS:
                to_delete.append(restriction.strip())
                fixed_lintian_tag(
                    'source', 'obsolete-runtime-tests-restriction',
                    '%s in line XX' % restriction.strip())
        if to_delete:
            removed_restrictions.extend(to_delete)
            paragraph['Restrictions'] = delete_from_list(
                paragraph['Restrictions'], to_delete)
            if not paragraph['Restrictions'].strip():
                del paragraph['Restrictions']


if 'needs-recommends' in removed_restrictions:
    # This is Certainty: possible, since the package may actually rely on the
    # (direct? indirect?) recommends, in which case we'd want to add them to
    # Depends.
    certainty = 'possible'
else:
    certainty = 'certain'


report_result(
    'Drop deprecated restriction{} {}. See '
    'https://salsa.debian.org/ci-team/autopkgtest/tree/'
    'master/doc/README.package-tests.rst'.format(
       's' if len(removed_restrictions) > 1 else '',
       ', ' .join(removed_restrictions)),
    certainty=certainty)
